! Calibrating the parameters of the directed search model.
!   For a detailed discussion, please refer to the main text.
module params
    implicit none
! General
    real*8, parameter :: one=1.0D+0
    real*8, parameter :: zero=0.0D+0
! Grids
    integer, parameter :: npts=250 ! Firm size
    integer, parameter :: nepspts=25 ! Productivity
    integer, parameter :: fnpts=1500 ! Finer grid size
    real*8, dimension (npts) :: ngrid
    real*8, dimension (nepspts) :: epsgrid
    real*8, dimension (fnpts) :: fngrid
    real*8, parameter :: nmin=zero ! Minimum firm size
    real*8, parameter :: nmax=50000.0D+0 ! Maximum firm size
    real*8, dimension (nepspts,nepspts) :: peps ! Transition matrix for productivity
    real*8, dimension (nepspts) :: pinv ! Invariant distribution over productivity

! Cost Function parameters
    real*8, parameter :: z1=3.0D+0 ! Qubic cost
    real*8, parameter :: z2=z1-one ! Model implication

! Other model parameters (calibrated)
    real*8, parameter :: alpha=0.677D+0 ! Fixed, RBC literature
    real*8, parameter :: rho=0.590D+0 ! Fixed, Abraham and White (2006)
    real*8, parameter :: beta=one/1.012D+00**(one/12.0D+00) ! Discount factor: quarterly interest rate
    real*8, parameter :: delta=0.015D+00/5.0D+00/4.0D+0 ! Exit prob: 1/6 of separations
    real*8, parameter :: lambda=0.0118D+0/4.0D+0 ! Seperation at zero growth rate (replacement workers)
    real*8, parameter :: psi=(one-delta)*(one-lambda)
    real*8, parameter :: b=0.94719736D+0 ! Normalization (from the WS calibration)

! Approximation error tolerance
    real*8, parameter :: tol=1.0D-6
    real*8, parameter :: mtol=1.0D-8
    real*8, parameter :: dtol=1.0D-8
    real*8, parameter :: gtol=1.0D-8


! Policy parameters
    real*8, parameter :: s=0.0D-0
    real*8, parameter :: tau=0.0D-0

! Calibrated parameters and calibration targets
    real*8 :: cv, cs, atech, sig, mu !, l
    real*8, parameter :: mmint=0.1D+0/4.1D+0 !Job filling probability at the lowest turnover group/zero employment growth
    real*8, parameter :: mmax=0.99999999999D+0 ! An upper bound on job filling probability
    real*8, parameter :: tsz=20.0D+0 ! firm size
    real*8, parameter :: tbr=0.73D+00 ! ratio of b to average Y
    real*8, parameter :: tbw=0.94455720D+00 ! ratio of b to average wage
    real*8, parameter :: thr=0.034D+00 ! hires rate (monthly)
    real*8, parameter :: tvy=1.3D+0/5.3D+0 ! job fillling probability (weekly)
    real*8, parameter :: ttheta=(one-(one-0.45D+00)**0.25D+0)/tvy ! matching function restriction
    real*8, parameter :: tels=0.28D+0 ! elasticity of job finding prob to market tightness
    real*8, parameter :: l=log(tels)/log(tvy)
    real*8, parameter :: xk=(one/tels-one)/ttheta**l ! scale parameter in the matching function

! Auxilary varibles
    real*8 :: gdiff, gdiff2, curdlt, curn, mmin
    integer :: gcount, dcount, cureps

! Policy functions
    real*8, dimension(npts,nepspts) :: g_dlt, g_env
    real*8, dimension(4,nepspts) :: g_ncr
    integer, dimension(4,nepspts) :: g_ncri, fg_ncri
    real*8, dimension(fnpts,nepspts) :: fg_dlt, fg_env, fg_np, fg_m, &
        fg_v, fg_r, fg_f, fg_dist, nwages, pwages, fg_j, fg_theta

! Cross-section variables
    integer, parameter :: ndpts= 1000000
    real*8, dimension(ndpts,6) :: datafw

end module params

program weekly_size_calib_rev
    use params
    implicit none
    integer :: j, npt, maxfun, iprint
    real*8 :: step, rhobeg, rhoend !, zbrent
    integer, parameter :: n=4
    real*8, dimension(n)  :: x, bl, bu
    real*8 :: w(1000)
!------------- Main program starts here----------!
! Options for bobyqa
    maxfun=1000
    iprint=2
    rhobeg=5.0D-2
    rhoend=1.0D-7
    npt=2*n+1

! Create grid points for firm size (ngrid)
ngrid(1)=nmin
ngrid(npts)=nmax
step=(ngrid(npts)-ngrid(1))/dble(real(npts-2))
do j=2,npts
   ngrid(j)=ngrid(j-1)+step
enddo
    g_dlt(:,:)=zero
    g_env(:,:)=zero
    g_ncr(1,:)=one
    g_ncr(2,:)=one+one
    g_ncr(3,:)=one+one+one

! Initialize policy functions and the envelope condition
!open(1,file='ngrid.txt',form='formatted')
!open(2,file='gdlt.txt',form='formatted')
!open(3,file='genv.txt',form='formatted')
!open(4,file='gncr.txt',form='formatted')
!do j=1,npts
!   read(1,'(f15.8)') ngrid(j)
!   read(2,'(25f15.8)') g_dlt(j,:)
!   read(3,'(25f15.8)') g_env(j,:)
!enddo
!read(4,'(25f15.8)') g_ncr(1,:)
!read(4,'(25f15.8)') g_ncr(2,:)
!read(4,'(25f15.8)') g_ncr(3,:)
!close(1)
!close(2)
!close(3)
!close(4)
    fg_dist(:,:)=zero
    fg_dist(1,1)=one

! Solve for the curvature parameter of the matching function. Its value targets ttheta
!    l=zbrent(rootl,1.0D-5,1.0D+1,tol)
print*, 'vacancy yield:', tvy, 'matching elasticity', l, 'matching scale', xk
!--Solve for the remaining parameters
! 1-cs, 2-sig, 3-atech, 4-mu
x(1)= 0.103D+0
x(2)= 0.217D+1
x(3)= 3.505D+0
x(4)= 0.052D+1

bl(1)=1.00D-4
bl(2)=1.00D-4
bl(3)=1.00D-2
bl(4)=1.00D-2

bu(1)=8.4D+4
bu(2)=9.9D+0
bu(3)=8.0D+0
bu(4)=2.8D+0

gcount=0
call bobyqa (n,npt,x,bl,bu,rhobeg,rhoend,iprint,maxfun,w)
call calfun(n,x,step)

! Step 5: Simulate cross section
    call monthly_simfw()
    call monthly_simfw_file()

    open(5,file='peps.txt',form='formatted')
    open(7,file='fngrid.txt',form='formatted')
    open(71,file='ngrid.txt',form='formatted')
    open(8,file='fgdlt.txt',form='formatted')
    open(81,file='gdlt.txt',form='formatted')
    open(9,file='fgdist.txt',form='formatted')
    open(91,file='genv.txt',form='formatted')
    open(92,file='gncr.txt',form='formatted')
    open(10,file='fgnp.txt',form='formatted')
    open(11,file='fgm.txt',form='formatted')
    open(12,file='fgv.txt',form='formatted')
    open(13,file='fgf.txt',form='formatted')
    open(131,file='fgr.txt',form='formatted')
    open(14,file='fgj.txt',form='formatted')
    open(15,file='fgwgsn.txt',form='formatted')
    open(16,file='fgwgsp.txt',form='formatted')
    do j=1,nepspts
       write(5,'(25f15.8)') peps(j,:), sum(peps(j,:))
    enddo
    write(5,'(25f15.8)') pinv(:), sum(pinv(:))
    do j=1,fnpts-1
       write(7,'(f15.8)') fngrid(j)
       write(8,'(25f15.8)') fg_dlt(j,:)
       write(9,'(25f15.8)') fg_dist(j,:)
       write(10,'(25f15.8)') fg_np(j,:)
       write(11,'(25f15.8)') fg_m(j,:)
       write(12,'(25f15.8)') fg_v(j,:)
       write(13,'(25f15.8)') fg_f(j,:)
       write(131,'(25f15.8)') fg_r(j,:)
       write(14,'(25f15.8)') fg_j(j,:)
       write(15,'(25f15.8)') nwages(j,:)
       write(16,'(25f15.8)') pwages(j,:)
    enddo
    do  j=1,npts
       write(71,'(f15.8)') ngrid(j)
       write(81,'(25f15.8)') g_dlt(j,:)
       write(91,'(25f15.8)') g_env(j,:)
    enddo
    write(92,'(25f15.8)') g_ncr(1,:)
    write(92,'(25f15.8)') g_ncr(2,:)
    write(92,'(25f15.8)') g_ncr(3,:)
    close(5)
    close(7)
    close(8)
    close(9)
    close(10)
    close(11)
    close(12)
    close(13)
    close(131)
    close(14)
    close(15)
    close(16)
    close(71)
    close(81)
    close(91)
    close(92)
    return
end program weekly_size_calib_rev

subroutine monthly_simfw()
    use params
    implicit none
    integer, parameter :: period=5
    integer, parameter :: ngrbin=194
    real*8, dimension(ngrbin+1) :: grbin
    real*8, dimension(ngrbin) :: grhrs, grseps, gruvcn, grvy, gremp
    integer, parameter :: nszbin=6
    real*8, dimension (nszbin+1) :: szbin
    real*8, dimension(nszbin) :: szhrs, szseps, szuvcn, szvy, szemp
    integer, parameter :: nwtbin=6
    real*8, parameter :: wt1=0.619D+0
    real*8, dimension (nwtbin) :: wtbin
    real*8, dimension(nwtbin) :: wthrs, wtseps, wtuvcn, wtvy, wtemp
    real*8, dimension(period) :: hrs, seps
    integer :: j, k, jlo
    real*8 :: v0, n0, wgt, hrm, sepm, np, gr, w, emp, &
       minsz, maxsz, empt, empb
    real*8, dimension(:,:), allocatable :: dataw
    real*8, dimension(:), allocatable :: datawt
    integer, dimension (:), allocatable :: op


    print*, 'simulating...'
    dcount=0
    do j=1,nepspts
    do k=1,fnpts
       !print*, k, j
       n0=fngrid(k)
       w=fg_dist(k,j)
       call sim_fw(1 ,period, j, n0, w, hrs, seps, n0, v0)
    enddo
    enddo
    allocate(dataw(dcount,6))
    dataw(:,:)=datafw(1:dcount,:)
    print*, 'work data created...'
    print*, 'vy-gr relationship: calculating...'
    grhrs(:)=zero
    grseps(:)=zero
    gruvcn(:)=zero
    gremp(:)=zero
    open(42,file='grbin.txt',form='formatted')
    do j=1,ngrbin+1
       read(42,'(f10.4)') grbin(j)
    enddo
    close(42)
    do j=1, dcount
        w=dataw(j,1)
        hrm=dataw(j,2)
        sepm=dataw(j,3)
        v0=dataw(j,4)
        n0=dataw(j,5)
        np=dataw(j,6)
        emp=(n0+np)/2.0D+0
        gr=2.0D+0*(np-n0)/(np+n0)
        jlo=1
        call hunt(grbin,ngrbin+1,gr,jlo,wgt)
        grhrs(jlo)=grhrs(jlo)+w*hrm
        grseps(jlo)=grseps(jlo)+w*sepm
        gruvcn(jlo)=gruvcn(jlo)+w*v0
        gremp(jlo)=gremp(jlo)+w*emp
    enddo
    empt=sum(gremp)
    do j=1, ngrbin
       grvy(j)=grhrs(j)/gruvcn(j)
       grhrs(j)=grhrs(j)/gremp(j)
       grseps(j)=grseps(j)/gremp(j)
       gruvcn(j)=gruvcn(j)/gremp(j)
       gremp(j)=gremp(j)/empt
    enddo
    open(101,file='grvy.txt',form='formatted')
    do j=1,ngrbin
       write(101,'(6f10.5)') grbin(j), grhrs(j), grseps(j), gruvcn(j), grvy(j), gremp(j)
    enddo
    print*, 'vy-gr relationship: written on grvy.txt'
    print*, 'sz-gr relationship: calculating...'
    minsz=log(minval((dataw(:,5)+dataw(:,6))/2.0D+0))
    maxsz=log(maxval((dataw(:,5)+dataw(:,6))/2.0D+0))
    szbin(1)=minsz
    szbin(nszbin+1)=maxsz
    w=(maxsz-minsz)/dble(real(nszbin))
    do j=2,nszbin
       szbin(j)=szbin(j-1)+w
    enddo
    szbin=exp(szbin)
    szhrs(:)=zero
    szseps(:)=zero
    szuvcn(:)=zero
    szemp(:)=zero
    do j=1, dcount
        w=dataw(j,1)
        hrm=dataw(j,2)
        sepm=dataw(j,3)
        v0=dataw(j,4)
        n0=dataw(j,5)
        np=dataw(j,6)
        emp=(n0+np)/2.0D+0
        jlo=1
        call hunt(szbin,nszbin+1,emp,jlo,wgt)
        szhrs(jlo)=szhrs(jlo)+w*hrm
        szseps(jlo)=szseps(jlo)+w*sepm
        szuvcn(jlo)=szuvcn(jlo)+w*v0
        szemp(jlo)=szemp(jlo)+w*emp
    enddo
    do j=1, nszbin
       szvy(j)=szhrs(j)/szuvcn(j)
       szhrs(j)=szhrs(j)/szemp(j)
       szseps(j)=szseps(j)/szemp(j)
       szuvcn(j)=szuvcn(j)/szemp(j)
       szemp(j)=szemp(j)/empt
    enddo
    open(102,file='szvy.txt',form='formatted')
    do j=1,nszbin
       write(102,'(6f10.5)') log(szbin(j)), szhrs(j), szseps(j), szuvcn(j), szvy(j), szemp(j)
    enddo
    close(102)
    print*, 'vy-sz relationship: written on szvy.txt'
    print*, 'wt-gr relationship: calculating...'
    wtbin(1)=wt1
    w=(one-wtbin(1))/dble(real(nwtbin-1))
    do j=2,nwtbin
       wtbin(j)=wtbin(j-1)+w
    enddo
    wtbin(nwtbin)=one
    allocate(datawt(dcount))
    allocate(op(dcount))
    do j=1, dcount
        hrm=dataw(j,2)
        sepm=dataw(j,3)
        n0=dataw(j,5)
        np=dataw(j,6)
        datawt(j)=2.0D+0*(hrm+sepm)/(n0+np)
    enddo
    op=(/(j,j=1,dcount)/)
    call ssort(datawt,op,dcount)
    wthrs(:)=zero
    wtseps(:)=zero
    wtuvcn(:)=zero
    wtemp(:)=zero
    empb=zero
    jlo=1
    do j=1, dcount
        w=dataw(op(j),1)
        hrm=dataw(op(j),2)
        sepm=dataw(op(j),3)
        v0=dataw(op(j),4)
        n0=dataw(op(j),5)
        np=dataw(op(j),6)
        emp=(n0+np)/2.0D+0
        if ((empb+wtemp(jlo)+w*emp)/empt.gt.wtbin(jlo)) then
           empb=empb+wtemp(jlo)
           jlo=jlo+1
        endif
        wthrs(jlo)=wthrs(jlo)+w*hrm
        wtseps(jlo)=wtseps(jlo)+w*sepm
        wtuvcn(jlo)=wtuvcn(jlo)+w*v0
        wtemp(jlo)=wtemp(jlo)+w*emp
    enddo
    do j=1, nwtbin
       wtvy(j)=wthrs(j)/wtuvcn(j)
       wthrs(j)=wthrs(j)/wtemp(j)
       wtseps(j)=wtseps(j)/wtemp(j)
       wtuvcn(j)=wtuvcn(j)/wtemp(j)
       wtemp(j)=wtemp(j)/empt
    enddo
    open(103,file='wtvy.txt',form='formatted')
    do j=1,nwtbin
       write(103,'(6f10.5)') wtbin(j), wthrs(j), wtseps(j), wtuvcn(j), wtvy(j), wtemp(j)
    enddo
    close(103)
    deallocate(datawt)
    deallocate(op)
    print*, 'vy-wt relationship: written on wtvy.txt'
    deallocate(dataw)
    print*, 'simulation ended'
    return

end


recursive subroutine sim_fw(t,period, ceps, cn, &
                     dx, hrs, seps, n0, v0)
   use params
   implicit none
   integer, intent(in) :: t, period, ceps
   real*8, intent(in) :: cn, dx
   real*8, intent(inout) :: n0, v0
   integer :: jlo, k
   real*8, dimension (period), intent(inout) :: hrs, seps
   real*8 :: wgt, dxx, np

   if (dx.le.1.0D-7) return
   jlo=1
   call hunt(fngrid,fnpts,cn,jlo,wgt)
   if (t.eq.1) then
      v0=wgt*(one-fg_m(jlo,ceps))*fg_v(jlo,ceps) &
         +(one-wgt)*(one-fg_m(jlo+1,ceps))*fg_v(jlo+1,ceps)
      n0=wgt*fg_np(jlo,ceps)+(one-wgt)*fg_np(jlo+1,ceps)
   endif
   if(t.eq.period) then ! Calculate cross section vacancy yield
      ! Final period policy choice and productivity
      np=wgt*fg_np(jlo,ceps)+(one-wgt)*fg_np(jlo+1,ceps)
      hrs(t)=wgt*fg_r(jlo,ceps)+(one-wgt)*fg_r(jlo+1,ceps)
      seps(t)=lambda*cn+wgt*fg_f(jlo,ceps) &
             +(one-wgt)*fg_f(jlo+1,ceps)
      dcount=dcount+1
      datafw(dcount,:)=(/dx, sum(hrs(2:period)), sum(seps(2:period)), v0, n0, np/)
      return
   else
          ! Calculate the current policy values for (n,eps)
          np=wgt*fg_np(jlo,ceps)+(one-wgt)*fg_np(jlo+1,ceps)
          hrs(t)=wgt*fg_r(jlo,ceps)+(one-wgt)*fg_r(jlo+1,ceps)
          seps(t)=lambda*cn+wgt*fg_f(jlo,ceps) &
                  +(one-wgt)*fg_f(jlo+1,ceps)
          do k=1,nepspts
             dxx=dx*peps(ceps,k)
             call sim_fw(t+1,period, k, np, &
                  dxx, hrs, seps, n0, v0)
          enddo
   endif
end

subroutine monthly_simfw_file()
    use params
    implicit none
    integer :: j, k
    integer, parameter:: period=5
    real*8, dimension(period) :: n, dlt, m
    integer, dimension(period) :: eps

    print*, 'Writing in file...'
    open (201,file='monthly_fw.txt',form='formatted')
    do j=1,nepspts
    do k=1,fnpts
       call sim_fw_file(1 ,period, j, fngrid(k), &
                    fg_dist(k,j), n,dlt,m,eps)
    enddo
    enddo
    close(201)
    print*, 'Written on file'

end

recursive subroutine sim_fw_file(t,period, ceps, cn, &
                     dx, n, dlt, m, eps)
   use params
   implicit none
   integer, intent(in) :: t, period, ceps
   real*8, intent(in) :: cn, dx
   integer :: jlo, k
   real*8, dimension (period), intent(inout) :: n, dlt, m
   integer, dimension (period), intent(inout) :: eps
   real*8 :: wgt, dxx

   if (dx.le.1.0D-7) return
   jlo=1
   call hunt(fngrid,fnpts,cn,jlo,wgt)
   if(t.eq.period) then ! Write in the file the simulated data (eps,n,dlt,p)
      ! Final period policy choice and productivity
          n(t)=cn
          eps(t)=ceps
          dlt(t)=wgt*fg_dlt(jlo,ceps)+(one-wgt)*fg_dlt(jlo+1,ceps)
          m(t)=wgt*fg_m(jlo,ceps)+(one-wgt)*fg_m(jlo+1,ceps)
      write(201,100) dx, eps, n, dlt, m
100   format(f25.16,5i3,15f15.8)
      return
   else
          ! Calculate the current policy values for (n,eps)
          n(t)=cn
          eps(t)=ceps
          dlt(t)=wgt*fg_dlt(jlo,ceps)+(one-wgt)*fg_dlt(jlo+1,ceps)
          m(t)=wgt*fg_m(jlo,ceps)+(one-wgt)*fg_m(jlo+1,ceps)
          n(t+1)=wgt*fg_np(jlo,ceps)+(one-wgt)*fg_np(jlo+1,ceps)
          do k=1,nepspts
             dxx=dx*peps(eps(t),k)
             call sim_fw_file(t+1,period, k, n(t+1), &
                  dxx, n, dlt, m, eps)
          enddo
   endif
end


subroutine calfun(np,x,fj)
    use params
    implicit none
    integer, intent(in) :: np
    real*8, intent(in), dimension(np) :: x
    real*8, intent(out) :: fj
    real*8 :: chr, cvy, csz, ccost, cbr, cbw
    real*8, dimension(nepspts,nepspts) :: pit
    integer :: j
    integer :: time(8)

    call date_and_time(values=time)
! Step 1: Set the parameter values to be calibrated
! 1-cs, 2-sig, 3-atech, 4-mu
    cs=x(1)*1.0D+2
    sig=x(2)*1.0D-1
    atech=x(3)
    mu=x(4)*1.0D-1

! Step 2: Calibrate cv to match the maximum possible value of p
    cv= mu*xk**(one/l)*mmint**(one+l)/(one-mmint**l)**((one+l)/l)

! Step 3: Create the productivity grid and calculate the transition
!    matrix and the unconditional distribution of productivities.
   call tauchen (nepspts,epsgrid,peps,rho,sig)
   epsgrid =atech*dexp(epsgrid)
   peps=peps/52.0D+0
   do j=1,nepspts
      peps(j,j)=51.0D+0/52.0D+0+peps(j,j)
   enddo

   ! Correct for probability sum equals to 1
   do j=1,nepspts
      peps(j,j)=one-sum(peps(j,1:j-1))-sum(peps(j,j+1:nepspts))
   enddo
   pit = peps
   do j = 1,100000
      pit = matmul(pit,peps)
   end do
   pinv = pit(1,:)
   ! Correct for probability sum equals to 1
   pinv(nepspts/2)=one-sum(pinv(1:nepspts/2-1))-sum(pinv(nepspts/2+1:nepspts))
   ! Solve firm's problem and calculate stationary equilibrium
    call firm_opt()
    call calcdist(fg_dist)
   ! Calculate current value of targets
    call calc_ctar(cvy, ccost, csz, chr, cbr, cbw)
   ! Calculate sum of squared errors
    fj= 1.0D+6*(chr-thr)**2+1.0D+4*(cvy-tvy)**2  &
        +1.0D+0*(csz-tsz)**2+1.0D+4*(cbw-tbw)**2


   ! Write output
   gcount=gcount+1
   print '(a43)', '+++++++++++++++++++++++++++++++++++++++++++'
   print '(a10,i5,5x,a9, i2,a1,i2,a1,i2)', 'Iteration=', gcount
   print '(a11, i2,a1,i2,a1,i2)',  'Start Time:', time(5), ':', time(6), ':', time(7)
   call date_and_time(values=time)
   print '(a11, i2,a1,i2,a1,i2)', 'End Time:  ', time(5), ':', time(6), ':', time(7)
   print*, ' '
   print '(a4,f15.8)','SSE=', fj
   print '(a15,f15.8)','Firm_Opt error=', gdiff
   print '(a15,f15.8)','Dist sum=      ', gdiff2
   print*, ' '
   print '(a25)', 'Current parameter values:'
   write(6,12) ' cv=', cv
   write(6,12) ' cs=', cs
   write(6,12) 'sig=', sig
   write(6,12) 'atc=', atech
   write(6,12) '  b=', b
   write(6,12) ' mu=', mu
   print*, ' '
   print '(a58)', 'Moment       Current        Target         Diff           '
   print '(a58)', '-------------------------------------------'
   write(6,13) 'JobFill Rate:', cvy, tvy, 1.0D+2*(cvy-tvy)
   write(6,13) 'Average Size:', csz, tsz, 1.0D+0*(csz-tsz)
   write(6,13) 'Hiring Rate :', chr, thr, 1.0D+2*(chr-thr)
   write(6,13) 'b/w         :', cbw, tbw, 1.0D+2*(cbw-tbw)
   write(6,13) 'b/(Y/N)     :', cbr, tbr, 1.0D+2*(cbr-tbr)
   write(6,*) 'Cost of Hrng:', ccost
   print '(a43)', '+++++++++++++++++++++++++++++++++++++++++++'
   print*, ' '
12 format(a4,f15.8)
13 format(a13,3f15.8)
   return
end subroutine calfun

subroutine calc_ctar(cvy, ccost, csz, chr, cbr, cbw)
! This subroutine calculates the target moments from the model
   use params
   implicit none
   real*8, intent(out) :: chr, cvy, csz, ccost, cbr, cbw
   integer :: j, k
   real*8 :: c_tot, w_tot, h_tot, pro_emp, ybar, &
             e, n, m, v, f, r, np, wgt, uvcn


   ! Initialize current targets at zero current targets
   chr=zero ! Aggregate hiring rate
   csz=zero ! Average size
   c_tot=zero ! Total hiring costs (excludes payments to workers)
   uvcn=zero
   pro_emp=zero ! # of production workers
   ybar=zero ! Total Output
   w_tot=zero ! Average wages of productive workers (excludes recruitment wages)
   h_tot=zero

   do j=1,nepspts
      do k=1,fnpts
! Skip zero-weighted observations
        if (fg_dist(k,j).eq.zero) cycle
! Write in current state, optimal decisions, and
! weight in the stationary distribution
        e=epsgrid(j)
        n=fngrid(k)
        wgt=fg_dist(k,j)
        v=fg_v(k,j)
        m=fg_m(k,j)
        f=fg_f(k,j)
        r=fg_r(k,j)
        np=fg_np(k,j)
! Calculate targeted moments from the model
        chr=chr+wgt*m*v
        csz=csz+wgt*(np+n)/2.0D+00
        c_tot=c_tot+wgt*(cs*v**z1/z1 &
              /(one+(one-lambda)*n)**z2+cv*v)
        uvcn=uvcn+wgt*v*(one-m)
        w_tot=w_tot+12.0D+0*wgt*(nwages(k,j)*min((one-lambda)*n,np) &
                +pwages(k,j)*r)
        h_tot=h_tot+wgt*(min((one-lambda)*n,np)+r)
        ybar=ybar+wgt*e*fg_np(k,j)**alpha
        pro_emp=pro_emp+wgt*((one-lambda)*n+r-f)
      enddo
   enddo
! Calculate the final values of the target moments

    cvy=chr/uvcn/(one+chr/uvcn)  ! Aggregate job-filling rate
    c_tot=c_tot/chr ! Total cost per hire
    chr=4.0D+0*chr/csz ! Aggregate hiring rate
    w_tot=w_tot/h_tot ! Average wages for production workers
    ccost=c_tot/w_tot ! Ratio of total recruitment cost
                      !    to aggregate production wages
    cbr=b/(ybar/pro_emp) ! Ratio of b to output per (employed) worker.
    cbw=12.0D+0*b/w_tot
    return
end

subroutine firm_opt()
   use params
   implicit none
   integer, parameter :: kmax=500
   integer :: j, k, kkk, jlo, t
   real*8, dimension (npts) :: ngrid0
   integer, dimension (npts) :: op
   integer, dimension (fnpts) :: fop
   real*8 :: step, zbrent, wgt, m, envfunc, diff, ediff, &
             ndiff, ddiff
   real*8, dimension(npts,nepspts) :: g_dlt0, g_env0
   real*8, dimension(4,nepspts) :: g_ncr0
    interface
       double precision function root1(x)
         implicit none
         real*8, intent(in) :: x
       end function root1
    end interface
    interface
       double precision function root2(x)
         implicit none
         real*8, intent(in) :: x
       end function root2
    end interface
    interface
       double precision function root3(x)
         implicit none
         real*8, intent(in) :: x
       end function root3
    end interface
    interface
       double precision function rootd(x)
         implicit none
         real*8, intent(in) :: x
       end function rootd
    end interface
    interface
       double precision function rootm(x)
         implicit none
         real*8, intent(in) :: x
       end function rootm
    end interface

mmin=mmint/2.0D+0
! Guess policy functions
g_env0=g_env
g_dlt0=g_dlt
g_ncr0=g_ncr
ngrid0=ngrid
! Iterate policy functions
do kkk=1,kmax
! Step 1: Update policy functions with the new solution
g_env=g_env0
g_dlt=g_dlt0
g_ncr=g_ncr0
ngrid=ngrid0
! Step 2: Solve for critical n values
!     i- Find the largest size firm with gdlt=0,
!        given epsilon
do j=1,nepspts
cureps=j
g_ncr(3,j)=zbrent(root3,nmin,nmax,gtol)
g_ncr(4,j)=(one-lambda)*g_ncr(3,j)
enddo
!    ii- Find the smallest size firm with gdlt=0,
!        given epsilon
do j=1,nepspts
cureps=j
g_ncr(2,j)=zbrent(root2,nmin,g_ncr(3,j)+1.0D+0,gtol)
enddo
!   iii- Find the firm size with gdlt=lambda*n,
!        given epsilon
do j=1,nepspts
cureps=j
g_ncr(1,j)=zbrent(root1,nmin,g_ncr(2,j)+1.0D+0,gtol)
enddo
! Step 4: Create a new grid (ngrid0) for firm size
!     This grid now includes all the critical values
!     Also calculate the location of these points
!     to be used in profit maximization
ngrid0(1)=nmin
ngrid0(npts)=nmax
ngrid0(npts-4*nepspts-1)=g_ncr(3,nepspts)+1.0D-0
step=2.0D+0*(ngrid0(npts-4*nepspts-1)-nmin)/dble(real(npts-4*nepspts-2))/dble(real(npts-4*nepspts-1))
do j=2,npts-4*nepspts-1
   ngrid0(j)=ngrid0(j-1)+dble(real(j-1))*step
enddo
t=npts-4*nepspts
   do j=1,4
      do k=1,nepspts
         ngrid0(t)=g_ncr(j,k)
         t=t+1
      enddo
   enddo
op=(/(j,j=1,npts)/)
call ssort(ngrid0,op,npts)
do j=1,4
   do k=1,nepspts
      jlo=1
      call hunt(ngrid0,npts,g_ncr(j,k),jlo,wgt)
      if (wgt.lt.tol) then
         g_ncri(j,k)=jlo+1
      else
         g_ncri(j,k)=jlo
      endif
   enddo
enddo
! Step 6: Solve profit maximization for hiring firms
!         and solve for the envelope condition
do j=1,nepspts
cureps=j
! Hiring firms
do k=1,g_ncri(2,j)-1
curn=ngrid0(k)
g_dlt0(k,j)=zbrent(rootd,zero,g_ncr(2,j)+1.0D+0,gtol)
curdlt=g_dlt0(k,j)
m=zbrent(rootm,mmin,mmax,mtol)
g_env0(k,j)=envfunc(m,ngrid0(k),g_dlt0(k,j),j)
enddo
! Adjustment for first cut-off
curn=ngrid0(g_ncri(1,j))
g_dlt0(g_ncri(1,j),j)=lambda*curn
curdlt=g_dlt0(g_ncri(1,j),j)
m=zbrent(rootm,mmin,mmax,mtol)
g_env0(g_ncri(1,j),j)=envfunc(m,curn,g_dlt0(g_ncri(1,j),j),j)
! Second cut-off to third cut-off (including 4th critical point)
do k=g_ncri(2,j),g_ncri(3,j)-1
curn=ngrid0(k)
g_dlt0(k,j)=zero
m=mmint
g_env0(k,j)=envfunc(m,curn,g_dlt0(k,j),j)
enddo
! Adjustment for third cut-off
g_dlt0(g_ncri(3,j),j)=zero
g_env0(g_ncri(3,j),j)=envfunc(m,ngrid0(k),-one,j)
! Third cut-off onward
do k=g_ncri(3,j)+1,npts
g_dlt0(k,j)=(one-lambda)*(g_ncr(3,j)-ngrid0(k))
g_env0(k,j)=envfunc(m,ngrid0(k),g_dlt0(k,j),j)
enddo
enddo
! Check for convergence
ndiff=maxval(abs(ngrid-ngrid0))
ddiff=maxval(abs(g_dlt-g_dlt0))
ediff=maxval(abs(g_env-g_env0))
diff=max(ndiff,ddiff,ediff)
if (diff.le.tol) exit
! End of iteration
enddo
gdiff=diff
g_dlt=g_dlt0
g_env=g_env0
ngrid=ngrid0
! Step 7: Create a finer grid
fngrid(1)=nmin
fngrid(fnpts)=nmax
fngrid(fnpts-4*nepspts-1)=g_ncr(4,nepspts)+1.0D-0
step=2.0D+0*(fngrid(fnpts-4*nepspts-1)-nmin)/dble(real(fnpts-4*nepspts-2))/dble(real(fnpts-4*nepspts-1))
do j=2,fnpts-4*nepspts-1
   fngrid(j)=fngrid(j-1)+dble(real(j-1))*step
enddo
t=fnpts-4*nepspts
   do j=1,4
      do k=1,nepspts
         fngrid(t)=g_ncr(j,k)
         t=t+1
      enddo
   enddo
fop=(/(j,j=1,fnpts)/)
call ssort(fngrid,fop,fnpts)
do j=1,4
   do k=1,nepspts
      jlo=1
      call hunt(fngrid,fnpts,g_ncr(j,k),jlo,wgt)
      if (wgt.lt.tol) then
         fg_ncri(j,k)=jlo+1
      else
         fg_ncri(j,k)=jlo
      endif
   enddo
enddo

! Step 8: Carry the policy functions over the finer grid

do j=1,nepspts
! Write in hiring firms decision
t=g_ncri(2,j)
do k=1,fg_ncri(2,j)-1
jlo=1
call hunt(ngrid(1:t),t,fngrid(k),jlo,wgt)
fg_dlt(k,j)=wgt*g_dlt(jlo,j)+(one-wgt)*g_dlt(jlo+1,j)
fg_env(k,j)=wgt*g_env(jlo,j)+(one-wgt)*g_env(jlo+1,j)
enddo
! correct no growth firms
fg_dlt(fg_ncri(1,j),j)=lambda*fngrid(fg_ncri(1,j))
fg_env(fg_ncri(1,j),j)=g_env(g_ncri(1,j),j)
! Shrinking firms
t=g_ncri(3,j)
do k=fg_ncri(2,j),fg_ncri(3,j)-1
jlo=1
call hunt(ngrid(1:t),t,fngrid(k),jlo,wgt)
fg_dlt(k,j)=zero
fg_env(k,j)=wgt*g_env(jlo,j)+(one-wgt)*g_env(jlo+1,j)
enddo
! Adjustment for the third cut-off
fg_dlt(fg_ncri(3,j),j)=zero
fg_env(fg_ncri(3,j),j)=envfunc(m,fngrid(k),-one,j)
do k=fg_ncri(3,j)+1,fnpts
fg_dlt(k,j)=(one-lambda)*(fngrid(fg_ncri(3,j))-fngrid(k))
fg_env(k,j)=envfunc(m,fngrid(k),fg_dlt(k,j),j)
enddo
enddo
! Write in other policy functions and workers value function
do j=1,nepspts
cureps=j
do k=1,fnpts
curn=fngrid(k)
curdlt=fg_dlt(k,j)
fg_np(k,j)=(one-lambda)*curn+fg_dlt(k,j)
if(k.ge.fg_ncri(2,j).and.k.le.fg_ncri(3,j)) fg_np(k,j)=(one-lambda)*curn
if (fg_dlt(k,j).gt.zero) then
fg_m(k,j)=zbrent(rootm,mmin,mmax,mtol)
fg_theta(k,j)=fg_m(k,j)*(xk/(one-fg_m(k,j)**l))**(one/l)
nwages(k,j)=b+mu
pwages(k,j)=fg_theta(k,j)*mu/fg_m(k,j)+b
fg_v(k,j)=fg_dlt(k,j)/fg_m(k,j)
fg_r(k,j)=fg_dlt(k,j)
fg_f(k,j)=zero
else
fg_m(k,j)=mmint
fg_theta(k,j)=zero
nwages(k,j)=b+mu
pwages(k,j)=b+mu
fg_v(k,j)=zero
fg_r(k,j)=zero
fg_f(k,j)=-fg_dlt(k,j)
endif

enddo
enddo

return
end subroutine firm_opt

subroutine calcdist(dist1)
    use params
    implicit none
    real*8, dimension(fnpts,nepspts), intent(inout) :: dist1
    real*8, dimension(fnpts,nepspts) :: dist0
    integer :: t, j, k, jj, jlo
    integer, parameter :: tmax=3000
    real*8 :: wgt, diff

    do t=1,tmax
    dist0=dist1
    dist1(:,:)=zero
       do j=1,nepspts
          dist1(1,j)=delta*pinv(j)
          do k=1,fnpts
             if (dist0(k,j).eq.zero) cycle
             jlo=1
             call hunt(fngrid,fnpts,fg_np(k,j),jlo,wgt)
             do jj=1,nepspts
                 dist1(jlo,jj)=dist1(jlo,jj)+peps(j,jj)*(one-delta)*dist0(k,j)*wgt
                 dist1(jlo+1,jj)=dist1(jlo+1,jj)+peps(j,jj)*(one-delta)*dist0(k,j)*(one-wgt)
             enddo
          enddo
       enddo
    diff=maxval(abs(dist1-dist0))
    if (diff.le.1.0D-10) exit
    enddo
    gdiff2=sum(dist1)
    return

end subroutine calcdist

double precision function root1(x)
    use params
    implicit none
    real*8, intent(in) :: x
    real*8 :: e, m, ej, wgt, denv
    real*8 :: zbrent
    integer :: j, jlo
    interface
       double precision function rootm(x)
         implicit none
         real*8, intent(in) :: x
       end function rootm
    end interface
    e=epsgrid(cureps)
    curdlt=lambda*x
    curn=x
    m=zbrent(rootm,mmin,mmax,mtol)
    ej=zero
    jlo=1
    call hunt(ngrid,npts,x,jlo,wgt)
    do j=1,nepspts
        denv=wgt*g_env(jlo,j)+(one-wgt)*g_env(jlo+1,j)
        ej=ej+peps(cureps,j)*denv
    enddo

    root1=alpha*e*x**(alpha-one)+beta*(one-delta)*ej &
          -b+s-mu*(xk/(one-m**l))**(one/l)/(one-m**l)
    return

end function root1

double precision function root2(x)
    use params
    implicit none
    real*8, intent(in) :: x
    real*8 :: e, m, ej, wgt, denv
    real*8 :: zbrent
    integer :: j, jlo
    interface
       double precision function rootm(x)
         implicit none
         real*8, intent(in) :: x
       end function rootm
    end interface

    e=epsgrid(cureps)
    curdlt=zero
    curn=x
    m=zbrent(rootm,mmint,mmax,mtol)
    ej=zero
    jlo=1
    call hunt(ngrid,npts,(one-lambda)*x,jlo,wgt)
    do j=1,nepspts
        denv=wgt*g_env(jlo,j)+(one-wgt)*g_env(jlo+1,j)
        ej=ej+peps(cureps,j)*denv
    enddo

    root2=alpha*e*((one-lambda)*x)**(alpha-one) &
          -b+s-mu*(xk/(one-m**l))**(one/l)/(one-m**l) &
          +beta*(one-delta)*ej
    return

end function root2


double precision function root3(x)
    use params
    implicit none
    real*8, intent(in) :: x
    real*8 :: e, ej, wgt, denv
    integer :: j, jlo

    e=epsgrid(cureps)
    ej=zero
    jlo=1
    call hunt(ngrid,npts,(one-lambda)*x,jlo,wgt)
    do j=1,nepspts
        denv=wgt*g_env(jlo,j)+(one-wgt)*g_env(jlo+1,j)
        ej=ej+peps(cureps,j)*denv
    enddo

    root3=alpha*e*((one-lambda)*x)**(alpha-one) &
          -b-mu+tau+beta*(one-delta)*ej
    return

end function root3

double precision function rootd(dx)
    use params
    implicit none
    real*8, intent(in) :: dx
    real*8 :: e, np, m, ej, wgt, denv
    real*8 :: zbrent
    integer :: j, jlo
    interface
       double precision function rootm(x)
         implicit none
         real*8, intent(in) :: x
       end function rootm
    end interface

    e=epsgrid(cureps)
    curdlt=dx
    np=(one-lambda)*curn+dx
    m=zbrent(rootm,mmin,mmax,mtol)
    ej=zero
    jlo=1
    call hunt(ngrid,npts,np,jlo,wgt)
    do j=1,nepspts
        denv=wgt*g_env(jlo,j)+(one-wgt)*g_env(jlo+1,j)
        ej=ej+peps(cureps,j)*denv
    enddo

    rootd=alpha*e*np**(alpha-one)+beta*(one-delta)*ej &
          -b+s-mu*(xk/(one-m**l))**(one/l)/(one-m**l)

    return

end function rootd


double precision function envfunc(m,x,dx,eps)
! Envelope(n, Dn,eps)
    use params
    implicit none
    integer, intent(in) :: eps
    real*8, intent(in) :: m,x,dx
    real*8 :: ej, wgt, denv, np
    integer :: j, jlo

    np=(one-lambda)*x+dx

    if (dx.lt.zero) then
        envfunc=-tau
    else
    ej=zero
    jlo=1
    call hunt(ngrid,npts,np,jlo,wgt)
    do j=1,nepspts
        denv=wgt*g_env(jlo,j)+(one-wgt)*g_env(jlo+1,j)
        ej=ej+peps(cureps,j)*denv
    enddo
    envfunc=alpha*epsgrid(eps)*np**(alpha-one)+beta*(one-delta)*ej &
            -b-mu+z2*cs*(dx/m)**z1*(one+(one-lambda)*x)**(-z2-one)/z1
    endif
    envfunc=(one-lambda)*envfunc
    return
end function envfunc

double precision function rootm(x)
   use params
   implicit none
   real*8, intent(in) :: x

   rootm=-mu*xk**(one/l)*x**(one+l)/(one-x**l)**((one+l)/l) &
         +cs*(curdlt/x)**(z1-one)*(one+(one-lambda)*curn)**(-z2)+cv

   return
end function rootm


      double precision function zbrent(func,x1,x2,tol)
      implicit none
      real*8 :: x1,x2,tol,zip,fa,fb,a,b,c,fc,d,e,tol1,xm,s,p,q,r
      integer :: iter
      real*8, parameter :: eps = 3.0D-08
      integer, parameter :: itmax = 200
      interface
         double precision function func(x)
         implicit none
         real*8, intent(in) :: x
         end function func
      end interface

      a = x1
      b = x2
      fa = func(a)
      fb = func(b)
      zip = 0.0D+00

      if ((fa .gt. zip .and. fb .gt. zip) .or. (fa .lt. zip .and. fb .lt. zip)) then
         print*, fa, fb, x1, x2
         stop 'root must be bracated'
      endif

      c = b
      fc = fb
      do iter = 1,itmax
         if ((fb .gt. zip .and. fc .gt. zip) .or. (fb .lt. zip .and. fc .lt. zip)) then
            c = a
            fc = fa
            d = b-a
            e = d
         endif
         if (dabs(fc) .lt. dabs(fb)) then
            a = b
            b = c
            c = a
            fa = fb
            fb = fc
            fc = fa
         endif
         tol1 = 2.0D+00*eps*dabs(b)+0.5D+00*tol
         xm = 0.5D+00*(c-b)
         if (dabs(xm) .le. tol1 .or. fb .eq. zip) then
            zbrent = b
            return
         endif
         if (dabs(e) .ge. tol1 .and. dabs(fa) .gt. dabs(fb)) then
            s = fb/fa
            if (a .eq. c) then
               p = 2.0D+00*xm*s
               q = 1.0D+00-s
            else
               q = fa/fc
               r = fb/fc
               p = s*(2.0D+00*xm*q*(q-r)-(b-a)*(r-1.0D+00))
               q = (q-1.0D+00)*(r-1.0D+00)*(s-1.0D+00)
            endif
            if (p .gt. zip) q = -q
            p = dabs(p)
            if (2.0D+00*p .lt. dmin1(3.0D+00*xm*q-dabs(tol1*q),dabs(e*q))) then
               e = d
               d = p/q
            else
               d = xm
               e = d
            endif
         else
            d = xm
            e = d
         endif
         a = b
         fa = fb
         if (dabs(d) .gt. tol1) then
            b = b+d
         else
            b = b+sign(tol1,xm)
         endif
         fb = func(b)
      end do

      zbrent = b
      stop 'zbrent exceeding maximum iterations'
      return
      end


      subroutine hunt(xx,n,x,jlo,wgt)
      implicit none
      integer, intent(in) :: n
      real*8, intent(in), dimension(n) :: xx
      real*8, intent(in) :: x
      integer, intent(out) :: jlo
      real*8, intent(out) :: wgt
      integer :: jhi,inc,jm
      logical :: ascnd

      ascnd = xx(n) .ge. xx(1)
      if (jlo .le. 0 .or. jlo .gt. n) then
         jlo = 0
         jhi = n-1
         goto 3
      endif

      inc = 1

      if (x .ge. xx(jlo) .eqv. ascnd) then
 1       jhi = jlo+inc
         if (jhi .gt. n) then
            jhi = n+1
         elseif (x .ge. xx(jhi) .eqv. ascnd) then
            jlo = jhi
            inc = inc+inc
            goto 1
         endif
      else
         jhi = jlo
 2       jlo = jhi-inc
         if (jlo .lt. 1) then
            jlo = 0
         elseif (x .lt. xx(jlo) .eqv. ascnd) then
            jhi = jlo
            inc = inc+inc
            goto 2
         endif
      endif
 3    if (jhi-jlo .eq. 1) then
         if (x .eq. xx(n)) jlo = n-1
         if (x .eq. xx(1)) jlo = 1
         go to 4
      endif
      jm = (jhi+jlo)/2
      if (x .ge. xx(jm) .eqv. ascnd) then
         jlo = jm
      else
         jhi = jm
      endif
      goto 3

4 continue
      wgt=(xx(jlo+1)-x)/(xx(jlo+1)-xx(jlo))
      return
      end

subroutine tauchen(n,z,p,rho,sig)
    implicit none
    integer :: j, k
    real*8 :: step
    integer, intent(in) :: n
    real*8, intent(in) :: rho, sig
    real*8, dimension(n), intent(out) :: z
    real*8, dimension(n,n), intent(out) :: p
    real*8 :: mu, cm, zmax, zmin, alnorm
    logical :: up
! Construct grid points for z
    zmin = -3.0D+0*sig/dsqrt(1.0D+0-rho**2)
    zmax =  3.0D+0*sig/dsqrt(1.0D+0-rho**2)
    step = (zmax-zmin)/dble(real((n-1)))
    z(1) = zmin
    do j = 2, n
      z(j) = z(j-1)+step
    end do
    z(n) = zmax
! Construct probabilities
    up=.false.
    mu=0.0D+0
    do j = 1, n
        cm = mu + rho*z(j)
        p(j,1) = alnorm((z(1) - cm + 0.5D+0*step)/sig,up)
        do k = 2, n-1
            p(j,k) = alnorm((z(k) - cm + 0.5D+0*step)/sig,up)  &
                     - alnorm((z(k) - cm - 0.5D+0*step)/sig,up)
        end do
        p(j,n) = 1.0D+0 - alnorm((z(n) - cm - 0.5D+0*step)/sig,up)
    end do

    return

end subroutine tauchen

FUNCTION alnorm( x, upper ) RESULT( fn_val )
!  Algorithm AS66 Applied Statistics (1973) vol.22, no.3

!  Evaluates the tail area of the standardised normal curve
!  from x to infinity if upper is .true. or
!  from minus infinity to x if upper is .false.

! ELF90-compatible version by Alan Miller
! Latest revision - 29 November 2001

   IMPLICIT NONE
   INTEGER, PARAMETER     :: dp = SELECTED_REAL_KIND(15, 100)
   REAL(DP), INTENT(IN)   ::  x
   LOGICAL,   INTENT(IN)  ::  upper
   REAL(DP)               ::  fn_val

   !  Local variables
   REAL(DP), PARAMETER   ::  zero=0.0_DP, one=1.0_DP, half=0.5_DP, con=1.28_DP
   REAL(DP)              ::  z, y
   LOGICAL               ::  up

   !  Machine dependent constants
   REAL(DP), PARAMETER  ::  ltone = 7.0_DP, utzero = 18.66_DP
   REAL(DP), PARAMETER  ::  p = 0.398942280444_DP, q = 0.39990348504_DP,   &
                            r = 0.398942280385_DP, a1 = 5.75885480458_DP,  &
                            a2 = 2.62433121679_DP, a3 = 5.92885724438_DP,  &
                            b1 = -29.8213557807_DP, b2 = 48.6959930692_DP, &
                            c1 = -3.8052E-8_DP, c2 = 3.98064794E-4_DP,     &
                            c3 = -0.151679116635_DP, c4 = 4.8385912808_DP, &
                            c5 = 0.742380924027_DP, c6 = 3.99019417011_DP, &
                            d1 = 1.00000615302_DP, d2 = 1.98615381364_DP,  &
                            d3 = 5.29330324926_DP, d4 = -15.1508972451_DP, &
                            d5 = 30.789933034_DP

   up = upper
   z = x
   IF( z < zero ) THEN
      up = .NOT. up
      z = -z
   END IF
   IF( z <= ltone  .OR.  (up  .AND.  z <= utzero) ) THEN
      y = half*z*z
      IF( z > con ) THEN
         fn_val = r*EXP( -y )/(z+c1+d1/(z+c2+d2/(z+c3+d3/(z+c4+d4/(z+c5+d5/(z+c6))))))
      ELSE
         fn_val = half - z*(p-q*y/(y+a1+b1/(y+a2+b2/(y+a3))))
      END IF
   ELSE
      fn_val = zero
   END IF

   IF( .NOT. up ) fn_val = one - fn_val
   RETURN
END FUNCTION alnorm
